import { parseISO } from 'date-fns'
import { ref, shallowReadonly, watch } from 'vue'
import { RouteLocationNormalizedLoaded, RouteParams, RouteRecordName } from 'vue-router'

import { useI18n } from '@/i18n'
import { useNavigationContext } from '@/stores/nav'
import { Show, TimeSlot } from '@/types'
import { fromAsync } from '@/util'

type Route = {
  name: RouteRecordName
  params?: RouteParams
}

type BreadcrumbLabel = { title: string }
type BreadcrumbLink = BreadcrumbLabel & ({ route: Route } | { link: string })
export type Breadcrumb = string | BreadcrumbLabel | BreadcrumbLink
export type BreadcrumbNormalized = { title: string; as: string; attrs: Record<string, unknown> }

type NavItemMenu = {
  title: string
  items: MenuItem[]
  requiresAdmin?: boolean
}

type NavItemLink = {
  route: Route
  label: string
  testId?: string
  requiresAdmin?: boolean
}

export type MenuItem = { menus?: NavItemMenu[] } & { link?: NavItemLink }

type MenuItemGeneratorOptions = Record<string, unknown> & {
  route: RouteLocationNormalizedLoaded
}

type MenuItemGenerator = (options: MenuItemGeneratorOptions) => AsyncIterable<MenuItem>

async function* getMainNavigation(options: MenuItemGeneratorOptions): AsyncIterable<MenuItem> {
  const { t } = useI18n()

  yield {
    link: {
      label: t('navigation.shows'),
      route: { name: 'shows' },
      testId: 'nav:shows',
    },
    menus: [...(await fromAsync(getShowContextMenu(options)))],
  }

  yield {
    link: {
      label: t('navigation.calendar'),
      route: { name: 'calendar' },
      testId: 'nav:calendar',
    },
  }
}

async function* getShowContextMenu(options: MenuItemGeneratorOptions): AsyncIterable<NavItemMenu> {
  if (!('show' in options) || !options.show) return
  const show = options.show as Show
  const { t } = useI18n()

  yield {
    title: show.name,
    items: [
      {
        link: {
          label: t('navigation.show.episodes'),
          route: {
            name: 'show-episodes',
            params: { ...options.route.params, showId: show.id.toString() },
          },
        },
        menus: [...(await fromAsync(getEpisodeContextMenu(options)))],
      },
      {
        link: {
          label: t('navigation.show.basicData'),
          route: {
            name: 'show-basic-data',
            params: { ...options.route.params, showId: show.id.toString() },
          },
        },
      },
    ],
  }
}

async function* getEpisodeContextMenu(
  options: MenuItemGeneratorOptions,
): AsyncIterable<NavItemMenu> {
  if (!('episode' in options) || !options.episode) return
  const episode = options.episode as TimeSlot
  const { t, locale } = useI18n()
  const start = parseISO(episode.start).toLocaleString(locale.value, {
    dateStyle: 'short',
    timeStyle: 'short',
  })

  yield {
    title: t('navigation.episode._title', { start }),
    items: [
      {
        link: {
          route: { name: 'show-episode-details', params: options.route.params },
          label: t('navigation.episode.details'),
        },
      },
    ],
  }
}

function useNavigationData(...menuCollectionGenerators: MenuItemGenerator[]) {
  const menuItemCollections = ref<MenuItem[]>([])
  const navigationContext = useNavigationContext()

  watch(
    navigationContext,
    async () => {
      const result: MenuItem[] = []
      for (const generator of menuCollectionGenerators) {
        try {
          const collections = generator(navigationContext.value)
          for await (const collection of collections) {
            result.push(collection)
          }
        } catch (e) {
          // pass
        }
      }
      menuItemCollections.value = result
    },
    { immediate: true },
  )

  return shallowReadonly(menuItemCollections)
}

export function useMainNavigationData() {
  return useNavigationData(getMainNavigation)
}
